/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_zero.h
 *
 * Purpose:     ZERO_v template class.
 *
 * Created:     29th July 2003
 * Updated:     15th January 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef STLSOFT_INCL_H_STLSOFT_ZERO
#define STLSOFT_INCL_H_STLSOFT_ZERO

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define _STLSOFT_VER_H_STLSOFT_ZERO_MAJOR      1
# define _STLSOFT_VER_H_STLSOFT_ZERO_MINOR      3
# define _STLSOFT_VER_H_STLSOFT_ZERO_REVISION   2
# define _STLSOFT_VER_H_STLSOFT_ZERO_EDIT       14
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef _STLSOFT_INCL_H_STLSOFT
 #include "stlsoft.h"   // Include the STLSoft root header
#endif /* !_STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// \brief Represents a type that can be an active replacement for ZERO
///
/// This class can act as a replacement for the ZERO macro, by being validly
/// assigned to or equated with pointer types only, as in
///
///   int   i = ZERO; // error
///   int   *p = ZERO; // OK
///
///   if(i == ZERO) {} // error
///   if(ZERO == i) {} // error
///
///   if(p == ZERO) {} // OK
///   if(ZERO == p) {} // OK
///
///
/// When used via inclusion of the file stlsoft_nulldef.h, the macro ZERO is
/// redefined as ZERO_v(), such that expressions containing ZERO will be valid
/// against pointers only.
struct ZERO_v
{
// Construction
public:
    /// Default constructor
    ZERO_v()
    {}

/// Static creation
public:
    static ZERO_v create()
    {
        return ZERO_v();
    }

// Conversion
public:
    /// A zero (of type ss_sint8_t)
    operator ss_sint8_t const () const
    {
        return 0;
    }
    /// A zero (of type ss_uint8_t)
    operator ss_uint8_t const () const
    {
        return 0;
    }
    /// A zero (of type ss_sint16_t)
    operator ss_sint16_t const () const
    {
        return 0;
    }
    /// A zero (of type ss_uint16_t)
    operator ss_uint16_t const () const
    {
        return 0;
    }
    /// A zero (of type ss_sint32_t)
    operator ss_sint32_t const () const
    {
        return 0;
    }
    /// A zero (of type ss_uint32_t)
    operator ss_uint32_t const () const
    {
        return 0;
    }
#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
    /// A zero (of type ss_sint64_t)
    operator ss_sint64_t const () const
    {
        return 0;
    }
    /// A zero (of type ss_uint64_t)
    operator ss_uint64_t const () const
    {
        return 0;
    }
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */

#ifdef __STLSOFT_CF_INT_DISTINCT_TYPE
    /// A zero (of type signed int)
    operator signed int const () const
    {
        return 0;
    }
    /// A zero (of type unsigned int)
    operator unsigned int const () const
    {
        return 0;
    }
#endif /* __STLSOFT_CF_INT_DISTINCT_TYPE */

    /// A zero (of type float)
    operator float const () const
    {
        return 0.0;
    }
    /// A zero (of type double)
    operator double const () const
    {
        return 0.0;
    }
    /// A zero (of type long double)
    operator long double const () const
    {
        return 0.0;
    }

// Not to be implemented
private:
    void operator &() const;

    ZERO_v(ZERO_v const &);
    ZERO_v const &operator =(ZERO_v const &);
};

/// operator == for ZERO_v and integral types
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_sint8_t i)         { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_uint8_t i)         { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_sint16_t i)        { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_uint16_t i)        { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_sint32_t i)        { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_uint32_t i)        { return i == 0; }
#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_sint64_t const &i) { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, ss_uint64_t const &i) { return i == 0; }
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */
#ifdef __STLSOFT_CF_INT_DISTINCT_TYPE
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, signed int i)         { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, unsigned int i)       { return i == 0; }
#endif /* __STLSOFT_CF_INT_DISTINCT_TYPE */
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, float const &i)       { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, double const &i)      { return i == 0; }
inline ss_bool_t operator ==(ZERO_v const &/* lhs */, long double const &i) { return i == 0; }

/// operator == for an integral type and ZERO_v
inline ss_bool_t operator ==(ss_sint8_t i, ZERO_v const &/* lhs */)         { return i == 0; }
inline ss_bool_t operator ==(ss_uint8_t i, ZERO_v const &/* lhs */)         { return i == 0; }
inline ss_bool_t operator ==(ss_sint16_t i, ZERO_v const &/* lhs */)        { return i == 0; }
inline ss_bool_t operator ==(ss_uint16_t i, ZERO_v const &/* lhs */)        { return i == 0; }
inline ss_bool_t operator ==(ss_sint32_t i, ZERO_v const &/* lhs */)        { return i == 0; }
inline ss_bool_t operator ==(ss_uint32_t i, ZERO_v const &/* lhs */)        { return i == 0; }
#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
inline ss_bool_t operator ==(ss_sint64_t const &i, ZERO_v const &/* lhs */) { return i == 0; }
inline ss_bool_t operator ==(ss_uint64_t const &i, ZERO_v const &/* lhs */) { return i == 0; }
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */
#ifdef __STLSOFT_CF_INT_DISTINCT_TYPE
inline ss_bool_t operator ==(signed int i, ZERO_v const &/* lhs */)         { return i == 0; }
inline ss_bool_t operator ==(unsigned int i, ZERO_v const &/* lhs */)       { return i == 0; }
#endif /* __STLSOFT_CF_INT_DISTINCT_TYPE */
inline ss_bool_t operator ==(float const &i, ZERO_v const &/* lhs */)       { return i == 0; }
inline ss_bool_t operator ==(double const &i, ZERO_v const &/* lhs */)      { return i == 0; }
inline ss_bool_t operator ==(long double const &i, ZERO_v const &/* lhs */) { return i == 0; }

/// operator != for ZERO_v and an integral type
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_sint8_t i)         { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_uint8_t i)         { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_sint16_t i)        { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_uint16_t i)        { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_sint32_t i)        { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_uint32_t i)        { return i != 0; }
#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_sint64_t const &i) { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, ss_uint64_t const &i) { return i != 0; }
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */
#ifdef __STLSOFT_CF_INT_DISTINCT_TYPE
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, signed int i)         { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, unsigned int i)       { return i != 0; }
#endif /* __STLSOFT_CF_INT_DISTINCT_TYPE */
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, float const &i)       { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, double const &i)      { return i != 0; }
inline ss_bool_t operator !=(ZERO_v const &/* lhs */, long double const &i) { return i != 0; }

/// operator != for an integral type and ZERO_v
inline ss_bool_t operator !=(ss_sint8_t i, ZERO_v const &/* lhs */)         { return i != 0; }
inline ss_bool_t operator !=(ss_uint8_t i, ZERO_v const &/* lhs */)         { return i != 0; }
inline ss_bool_t operator !=(ss_sint16_t i, ZERO_v const &/* lhs */)        { return i != 0; }
inline ss_bool_t operator !=(ss_uint16_t i, ZERO_v const &/* lhs */)        { return i != 0; }
inline ss_bool_t operator !=(ss_sint32_t i, ZERO_v const &/* lhs */)        { return i != 0; }
inline ss_bool_t operator !=(ss_uint32_t i, ZERO_v const &/* lhs */)        { return i != 0; }
#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
inline ss_bool_t operator !=(ss_sint64_t const &i, ZERO_v const &/* lhs */) { return i != 0; }
inline ss_bool_t operator !=(ss_uint64_t const &i, ZERO_v const &/* lhs */) { return i != 0; }
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */
#ifdef __STLSOFT_CF_INT_DISTINCT_TYPE
inline ss_bool_t operator !=(signed int i, ZERO_v const &/* lhs */)         { return i != 0; }
inline ss_bool_t operator !=(unsigned int i, ZERO_v const &/* lhs */)       { return i != 0; }
#endif /* __STLSOFT_CF_INT_DISTINCT_TYPE */
inline ss_bool_t operator !=(float const &i, ZERO_v const &/* lhs */)       { return i != 0; }
inline ss_bool_t operator !=(double const &i, ZERO_v const &/* lhs */)      { return i != 0; }
inline ss_bool_t operator !=(long double const &i, ZERO_v const &/* lhs */) { return i != 0; }

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_ZERO */

/* ////////////////////////////////////////////////////////////////////////// */
